package PVE::INotifyEtcnetOverride;

use strict;
use warnings;

use Clone qw(clone);

use IO::File;
use PVE::Tools;
use PVE::ProcFSTools;
use File::Basename;
use File::Path qw(make_path remove_tree);

no warnings qw(redefine);

our $die_on_global_config_lines=1;
our $tolerate_ovs_configuration=0;
our $create_pvedebug_dump_file=0;

#----------------------------------------------------
my $bond_modes = { 'balance-rr' => 0,
		   'active-backup' => 1,
		   'balance-xor' => 2,
		   'broadcast' => 3,
		   '802.3ad' => 4,
		   'balance-tlb' => 5,
		   'balance-alb' => 6,
	       };

my $ovs_bond_modes = {
    'active-backup' => 1,
    'balance-slb' => 1,
    'lacp-balance-slb' => 1,
    'lacp-balance-tcp' => 1, 
};

my $parse_ovs_option = sub {
    my ($data) = @_;

    my $opts = {};
    foreach my $kv (split (/\s+/, $data || '')) {
	my ($k, $v) = split('=', $kv, 2);
	$opts->{$k} = $v if $k && $v;
    }
    return $opts;
};

my $set_ovs_option = sub {
    my ($d, %params) = @_;

    my $opts = &$parse_ovs_option($d->{ovs_options});

    foreach my $k (keys %params) {
	my $v = $params{$k};
	if ($v) {
	    $opts->{$k} = $v;
	} else {
	    delete $opts->{$k};
	}
    }

    my $res = [];
    foreach my $k (keys %$opts) {
	push @$res, "$k=$opts->{$k}";
    }

    if (my $new = join(' ', @$res)) {
	$d->{ovs_options} = $new;
	return $d->{ovs_options};
    } else {
	delete $d->{ovs_options};
	return undef;
    }
};

my $extract_ovs_option = sub {
    my ($d, $name) = @_;

    my $opts = &$parse_ovs_option($d->{ovs_options});

    my $v = delete $opts->{$name};

    my $res = [];
    foreach my $k (keys %$opts) {
	push @$res, "$k=$opts->{$k}";
    }

    if (my $new = join(' ', @$res)) {
	$d->{ovs_options} = $new;
    } else {
	delete $d->{ovs_options};
    }

    return $v;
};

#----------------------------------------------------




# override PVE::INotify (to minimize upstream code editing)
sub PVE::INotify::read_etc_network_interfaces {
    my ($filename, $fh) = @_;
    my $proc_net_dev = IO::File->new('/proc/net/dev', 'r');
    my $active = PVE::ProcFSTools::get_active_network_interfaces();
    return PVE::INotify::__read_etc_network_interfaces($fh, $proc_net_dev, $active);
}
sub PVE::INotify::write_etc_network_interfaces {
    my ($filename, $fh, $config) = @_;
    my $raw = PVE::INotify::__write_etc_network_interfaces($config);
    PVE::Tools::safe_print($filename, $fh, $raw);
    __write_etc_net_interfaces($config);
}

sub read_etc_net_interfaces {
    my (undef) = @_;
    my $proc_net_dev = IO::File->new('/proc/net/dev', 'r');
    my $active = PVE::ProcFSTools::get_active_network_interfaces();
    return __read_etc_net_interfaces(undef, $proc_net_dev, $active);
}

sub __list_etcnet_interface_dirs {
    my $prefix=shift;
    $prefix ||= '/etc/net/ifaces';
    # remove etcnet interfaces not mentioned in PVE
    # lo is touched neither by the reader nor by the writer; 
    # default and unknown in /etc/net/ifaces are etcnet specific
    my @out;
    foreach my $tainted_path (glob($prefix.'/*')) {
	if ($tainted_path =~ m!^($prefix/(.*))$!) {
	    my $path = $1;
	    my $iface = $2;
	    push @out, $path if -d $path and $iface ne 'default' and $iface ne 'unknown' and $iface ne 'lo';
	} else {
	    die "Oops: can't happen: bad tainted_path $tainted_path";
	}
    }
    return @out;
}

sub __write_to_file {
    my ($path, $data)=@_;
    open (my $fd, '>', $path) or die "can't open $path: $!";
    print $fd $data;
    close ($fd) or die "can't close $path: $!";
}

sub __write_etc_net_interfaces {
    my ($config, $prefix) = @_;
    $prefix ||= '/etc/net/ifaces';
    my $raw = __write_etc_net_interfaces_prepare_tree($config);
    die "Oops! Etcnet not installed: $prefix not found" if ! -d $prefix;
    foreach my $iface (keys (%$raw)) {
	my $dir = $prefix.'/'.$iface;
	if (! -d $dir) {
	    die "mkdir $dir failed: $!" unless mkdir $dir, 0755;
	}
	foreach my $file (keys (%{$raw->{$iface}})) {
	    my $path = $dir.'/'.$file;
	    my $filebody = $raw->{$iface}->{$file};
	    if (defined $filebody) {
		__write_to_file($path, $filebody);
	    } elsif (-e $path) {
		die "unlink $path failed: $!" unless unlink $path;
	    }
	}
    }
    # remove etcnet interfaces not mentioned in PVE
    foreach my $ifacedir (__list_etcnet_interface_dirs($prefix)) {
	remove_tree ($ifacedir) unless $raw->{basename($ifacedir)};
    }
}

sub __write_etc_net_interfaces_as_string {
    my ($config) = @_;
    my $raw=__write_etc_net_interfaces_prepare_tree($config);
    my $str='';
    foreach my $iface (sort {$a cmp $b} keys (%$raw)) {
	foreach my $file (sort {$a cmp $b} keys (%{$raw->{$iface}})) {
	    my $filebody=$raw->{$iface}->{$file};
	    $str .= "/etc/net/ifaces/$iface/$file:\n".$filebody if defined $filebody;
	}
    }
    return $str;
}

sub __write_etc_net_interfaces_prepare_tree {
    my ($config) = @_;

    PVE::INotify::__prewrite_setup_network_interfaces($config);

    # support for tests in test/etc_network_interfaces
    my @options;
    foreach my $opt (@{$config->{options}}) {
	next if $opt->[1] =~ /^\s*source(?:-directory)?\s/;
	next if $opt->[1] =~ /^\s*allow-ovs/;
	push @options, $opt;
    }
    if (@options && $die_on_global_config_lines) {
	die "Etcnet writer: got /etc/network/interface global config lines: ",join('; ',map {$_->[0].','.$_->[1]} @options);
    }
    # end support for tests in test/etc_network_interfaces

    my $ifaces = $config->{ifaces};

    my $known_ovs_ports = {};
    foreach my $iface (keys %$ifaces) {
	my $type = $config->{$iface}->{type};
	$type ||= $config->{$iface}->{ovs_type};
	$known_ovs_ports->{$iface} = $type if $type and ($type eq 'OVSPort' or $type eq 'OVSIntPort');
    }

    my $fstree={};
    foreach my $iface (keys %$ifaces) {
	__interface_to_etcnet($fstree, $iface, $ifaces->{$iface}, $known_ovs_ports);
    }
    return $fstree;
}

sub __assert_set($$$) {
    my ($hash,$opt,$val)=@_;
    my $oldval=$hash->{$opt};
    die "__set: conflict for set $opt=$val: already ".$oldval if defined($oldval) and $oldval ne $val;
    $hash->{$opt}=$val;
}

sub __unique_array($) {
    my ($arrayref)=@_;
    my @out;
    my %check;
    for (my $i=0;$i<=$#{$arrayref};$i++) {
	my $curitem=$arrayref->[$i];
	push @out, $curitem unless $check{$curitem};
	$check{$curitem}=1;
    }
    return @out;
}

sub __etcnet_unquote($) {
    my $in=shift;
    return '' unless defined $in;
    $in=~s/^[\s'"]*//;
    $in=~s/[\s'"]*$//;
    return $in;
}

sub __etcnet_listify($) {
    my $in=shift;
    $in=~s/^[\s'"]*//;
    $in=~s/[\s'"]*$//;
    return "'$in'";
}

sub __etcnet_unlistify($) {
    my $in=shift;
    $in=~s/^[\s'"]*//;
    $in=~s/[\s'"]*$//;
    my $devs = {};
    foreach my $p (split (/\s+/, $in)) {
	next if $p eq 'none';
	$devs->{$p} = 1;
    }
    return join (' ', sort keys %{$devs});
}

sub __etcnet_netmask4($) {
    my $in=shift;
    if ($in eq '255.255.255.255') {
	return 32;
    } elsif ($in eq '255.255.255.254') {
	return 31;
    } elsif ($in eq '255.255.255.252') {
	return 30;
    } elsif ($in eq '255.255.255.248') {
	return 29;
    } elsif ($in eq '255.255.255.240') {
	return 28;
    } elsif ($in eq '255.255.255.224') {
	return 27;
    } elsif ($in eq '255.255.255.192') {
	return 26;
    } elsif ($in eq '255.255.255.128') {
	return 25;
    } elsif ($in eq '255.255.255.0') {
	return 24;
    } elsif ($in eq '255.255.254.0') {
	return 23;
    } elsif ($in eq '255.255.252.0') {
	return 22;
    } elsif ($in eq '255.255.248.0') {
	return 21;
    } elsif ($in eq '255.255.240.0') {
	return 20;
    } elsif ($in eq '255.255.224.0') {
	return 19;
    } elsif ($in eq '255.255.192.0') {
	return 18;
    } elsif ($in eq '255.255.128.0') {
	return 17;
    } elsif ($in eq '255.255.0.0') {
	return 16;
    } elsif ($in eq '255.254.0.0') {
	return 15;
    } elsif ($in eq '255.252.0.0') {
	return 14;
    } elsif ($in eq '255.248.0.0') {
	return 13;
    } elsif ($in eq '255.240.0.0') {
	return 12;
    } elsif ($in eq '255.224.0.0') {
	return 11;
    } elsif ($in eq '255.192.0.0') {
	return 10;
    } elsif ($in eq '255.128.0.0') {
	return 9;
    } elsif ($in eq '255.0.0.0') {
	return 8;
    } elsif ($in eq '254.0.0.0') {
	return 7;
    } elsif ($in eq '252.0.0.0') {
	return 6;
    } elsif ($in eq '248.0.0.0') {
	return 5;
    } elsif ($in eq '240.0.0.0') {
	return 4;
    } elsif ($in eq '224.0.0.0') {
	return 3;
    } elsif ($in eq '192.0.0.0') {
	return 2;
    } elsif ($in eq '128.0.0.0') {
	return 1;
    } elsif ($in eq '0.0.0.0') {
	return 0;
    }
    die "etcnet: unsupported netmask $in";
}

sub __etcnet_bondmode($) {
    my $mode=shift;
    return $mode if $mode =~ /^\d+$/;
    return $bond_modes->{$mode} if defined($bond_modes->{$mode});
    die "Oops! __etcnet_bondmode: unsupported BONDMODE $mode";
}

sub __etcnet_set_host($$$) {
    my ($options, $d, $param) = @_;
    my $val = $d->{$param};
    if (defined $val) {
	if ($val eq 'none' or $val eq '') {
	    $options->{'HOST'}='';
	} else {
	    $options->{'HOST'}=__etcnet_listify($val);
	}
    }
    delete $d->{$param};
}

##    $raw->{'options'} = <<'NETWORKDOC';
# network interface settings; autogenerated
# Please do NOT modify this file directly, unless you know what
# you're doing.
#
# If you want to manage part of the network configuration manually,
# please utilize the 'source' or 'source-directory' directives to do
# so.
# PVE will preserve these directives, but will NOT its network
# configuration from sourced files, so do not attempt to move any of
# the PVE managed interfaces into external files!
##TODO_PVE_MANAGED=yes
##NETWORKDOC

sub __interface_to_etcnet {
    my ($raws, $iface, $orig_d, $known_ovs_ports) = @_;
    return '' if ! $orig_d;
    # etcnet knows lo better
    return if $iface eq 'lo';

    my $d=clone($orig_d);
    my $raw= {
	'options' => undef,
	'PVEDEBUG' => undef,
	'brctl' => undef,
	'ipv4address' => undef,
	'ipv4route' => undef,
	'ipv6address' => undef,
	'ipv6route' => undef,

    };
    $raws->{$iface}=$raw;

    my $options={};

    my $families=$d->{families};
    if ($families) {
	foreach my $family (@$families) {
	    if ($family eq 'inet') {
		__assert_set($options,'CONFIG_IPV4','yes');
		die "Oops! family eq inet, method is expected" unless $d->{method};
	    } elsif ($family eq 'inet6') {
		__assert_set($options,'CONFIG_IPV6','yes');
		die "Oops! family eq inet6, method6 is expected" unless $d->{method6};
	    } else {
		die "Oops!: $iface has unknown family $family";
	    }
	}
	delete($d->{families});
    }

    my $bootproto='static';
    my $method=$d->{method};
    $bootproto=$method if $method and $method ne 'manual';
    __assert_set($options,'BOOTPROTO',$bootproto);
    delete $d->{method};
    delete $d->{method6};

    my @extopts;
    my $vlan_raw_device;
    if ($d->{options}) {
	foreach my $extopt (@{$d->{options}}) {
	    if ($extopt =~ /^(?:network|pre-up\s+ifup)\s/) {
		# ignore
	    } elsif ($extopt =~ /^pre-up\s+ifconfig\s+\S+\s+up/) {
		#pre-up ifconfig bond0 up
		# ignore
	    } elsif ($extopt =~ s/^vlan_raw_device\s*//) {
		die "Oops! multiple vlan_raw_devices found for $iface: $vlan_raw_device $extopt" if $vlan_raw_device;
		$vlan_raw_device=$extopt;
	    } else {
		push @extopts, $extopt;
	    }
	}
	$d->{options}=\@extopts;
	delete $d->{options} if ! scalar @extopts;
    }

    my @ovs_options;
    push @ovs_options, $d->{ovs_options} if $d->{ovs_options};
    delete $d->{ovs_options};

    my $type=$d->{type};
    my $ovstype=$d->{ovs_type};
    if (! defined($type) && $ovstype) {
	$type = $ovstype;
    }

    #iface vlan3 inet static
    #        address  0.0.0.0
    #        netmask  0.0.0.0
    #        vlan_raw_device bond2
    $type='vlan' if $type eq 'unknown' and $vlan_raw_device;
    
    if ($type eq 'eth') {
	$options->{'TYPE'}='eth';
	### TODO!!!
	$options->{CONFIG_WIRELESS}='no';
    } elsif ($type eq 'vlan') {
	$options->{'TYPE'}='vlan';
	$options->{'HOST'}=$vlan_raw_device;
	if ($iface =~ /(\d+)(?::\d+)?$/) { # id.VID:ALIAS
	    $options->{'VID'}=$1;
	}
	if ($vlan_raw_device =~ /^vmbr/) {
	    if ($iface =~ /^$vlan_raw_device\./) {
		$options->{'TYPE'}='brivlanport';
	    } else {
		die "Oops! etcnet writer: vlan brivlanport $iface: name expected to start from $vlan_raw_device";
	    }
	}
    } elsif ($type eq 'bond') {
	$options->{'TYPE'}='bond';
	__etcnet_set_host($options,$d,'slaves');
	my @bondoptions;
	my $v = defined ($d->{'bond_miimon'}) ? $d->{'bond_miimon'} : 100;
	push @bondoptions, "miimon=$v";
	delete $d->{'bond_miimon'};

	$v = defined ($d->{'bond_mode'}) ? $d->{'bond_mode'} : 'balance-rr';
	$options->{'BONDMODE'} = __etcnet_bondmode($v);

	if ($d->{'bond_mode'} && $d->{'bond_xmit_hash_policy'} &&
	    ($d->{'bond_mode'} eq 'balance-xor' || $d->{'bond_mode'} eq '802.3ad')) {
	    push @bondoptions, "xmit_hash_policy=$d->{'bond_xmit_hash_policy'}";
	}

	push @bondoptions, "lacp_rate=$d->{'lacp_rate'}" if defined $d->{'lacp_rate'};
	push @bondoptions, "downdelay=$d->{'downdelay'}" if defined $d->{'downdelay'};
	push @bondoptions, "updelay=$d->{'updelay'}" if defined $d->{'updelay'};
	#push @bondoptions, "=$d->{''}" if defined $d->{''};

	map {delete $d->{$_}} qw/bond_mode bond_xmit_hash_policy lacp_rate downdelay updelay/;

	__assert_set($options,'BOOTPROTO','static');
	$options->{'BONDOPTIONS'} = "'".join(' ',@bondoptions)."'" if @bondoptions;

    } elsif ($type eq 'bridge') {
	$options->{'TYPE'}='bri';
	__etcnet_set_host($options,$d,'bridge_ports');
	my $brctl='';
	my $v = defined($d->{bridge_stp}) ? $d->{bridge_stp} : 'off';
	$brctl .= "stp AUTO $v\n";
	delete $d->{bridge_stp};

	$v = defined($d->{bridge_fd}) ? $d->{bridge_fd} : 0;
	$brctl .= "setfd AUTO $v\n";
	delete $d->{bridge_fd};

	if(defined($d->{bridge_vlan_aware})) {
	    $options->{VLAN_AWARE}=$d->{bridge_vlan_aware};
	    delete $d->{bridge_vlan_aware};
	}
	if(defined($d->{bridge_vids})) {
	    $options->{VIDS}=$d->{bridge_vids};
	    delete $d->{bridge_vids};
	}
	if(defined($d->{bridge_pvid})) {
	    $options->{VIDS}=$d->{bridge_pvid};
	    delete $d->{bridge_pvid};
	}

	$raw->{'brctl'} = $brctl;

    } elsif ($type eq 'OVSBridge') {
	$options->{'TYPE'}='ovsbr';
	my $val = $d->{'ovs_ports'};
	if (!defined $val or $val eq 'none') {
	    die "Oops! etcnet writer: OVSBridge $iface: expected ovs_ports";
	} else {
	    $val =~ s/['"]//g;
	    # in etcnet we write only real interfaces
	    my @ovs_ports = grep {not $known_ovs_ports->{$_}} split(/\s+/,$val);
	    delete $d->{'ovs_ports'};
	    $d->{'ovs_ports'} = join(' ',@ovs_ports) if @ovs_ports;
	}
	__etcnet_set_host($options,$d,'ovs_ports');
	die "Oops! $iface: ovs_ports: set_host but still alive" if defined $d->{ovs_ports};
	die "Oops! ovs_type: expected $type, got $d->{ovs_type}" if defined $d->{ovs_type} and $d->{ovs_type} ne $type;
	delete($d->{ovs_type});

    } elsif ($type eq 'OVSPort' || $type eq 'OVSIntPort' || $type eq 'OVSBond') {

	die "Oops! ovs_type: expected $type, got $d->{ovs_type}" if defined $d->{ovs_type} and $d->{ovs_type} ne $type;
	delete $d->{ovs_type};

	$d->{autostart} = 0; # started by the bridge

	if (defined($d->{ovs_tag})) {
	    push @ovs_options, 'tag='.$d->{ovs_tag};
	}
	delete $d->{ovs_tag};

	if (defined $d->{ovs_bridge}) {
	    $options->{BRIDGE}=$d->{ovs_bridge};
	    delete($d->{ovs_bridge});
	}
	if (defined $d->{ovs_extra}) {
	    $options->{OVS_EXTRA}=$d->{ovs_extra};
	    delete($d->{ovs_extra});
	}

	if ($type eq 'OVSPort') {
	    $options->{'TYPE'}='ovsport';
	} elsif ($type eq 'OVSIntPort') {
  	    $options->{'TYPE'}='ovsport';
	} elsif ($type eq 'OVSBond') {
	    $options->{'TYPE'}='ovsbond';
	    $d->{bond_mode} = 'active-backup' if !$d->{bond_mode};
	    $ovs_bond_modes->{$d->{bond_mode}} ||
		die "OVS does not support bond mode '$d->{bond_mode}\n";

	    if ($d->{bond_mode} eq 'lacp-balance-slb') {
		push @ovs_options, 'lacp=active', 'bond_mode=balance-slb';
	    } elsif ($d->{bond_mode} eq 'lacp-balance-tcp') {
		push @ovs_options, 'lacp=active', 'bond_mode=balance-tcp';
	    } else {
		push @ovs_options, 'bond_mode='.$d->{bond_mode};
	    }
	    delete $d->{bond_mode};
	    __etcnet_set_host($options,$d,'ovs_bonds');
	}
    } else {
	use Data::Dumper;
	die "TODO: etcnet: $iface has unsupported type <$type>. Arguments: ",Dumper($orig_d);
    }
    delete($d->{type});

    $options->{'OVS_OPTIONS'}="'".join(' ', __unique_array(\@ovs_options) )."'" if @ovs_options;


    if (defined $d->{address}) {
	$raw->{'ipv4address'} = $d->{address}.'/'. __etcnet_netmask4($d->{netmask} // '255.255.255.0')."\n";
	__assert_set($options,'CONFIG_IPV4','yes');
	$raw->{'ipv4address'} .= 'broadcast '.$d->{broadcast}."\n" if $d->{broadcast};
	delete($d->{address});
	delete($d->{netmask});
	delete($d->{broadcast});
    }
    if (defined $d->{gateway}) {
	$raw->{'ipv4route'}='default via '.$d->{gateway}."\n";
	delete($d->{gateway});
    }
    if (defined $d->{address6}) {
	$raw->{'ipv6address'} = $d->{address6}.'/'. ($d->{netmask6} // 112)."\n";
	__assert_set($options,'CONFIG_IPV6','yes');
	$raw->{'ipv6address'} .= 'broadcast '.$d->{broadcast6}."\n" if $d->{broadcast6};
	delete($d->{address6});
	delete($d->{netmask6});
	delete($d->{broadcast6});
    }
    if (defined $d->{gateway6}) {
	$raw->{'ipv6route'}='default via '.$d->{gateway6}."\n";
	__assert_set($options,'CONFIG_IPV6','yes');
	delete($d->{gateway6});
    }

    $options->{'ONBOOT'}='yes' if $d->{autostart};
    delete $d->{autostart};

    if ($options) {
	my $optstring='';
	foreach my $key (sort {$a cmp $b} keys %$options) {
	    $optstring .= $key . '=' . $options->{$key} . "\n";
	}
	foreach my $option (@{$d->{options}}) {
	    $optstring .= $option."\n";
	}

	# add comments
	my $comments = $d->{comments} || '';
	foreach my $cl (split(/\n/, $comments)) {
	    $optstring .= "#$cl\n";
	}
	$raw->{'options'}=$optstring;
    }

    map {delete $d->{$_}} qw/priority active exists comments comments6 options options6/;

    if (%$d) {
	require Data::Dumper;
	if ($create_pvedebug_dump_file) {
	    #print Dumper($d) if %$d;
	    $raw->{'PVEDEBUG'}="The following options are not handled:\n".Dumper($d);
	} else {
	    die "Etcnet writer: $iface ($type): The following options are not handled: ".Dumper($d);
	}
    }
}

sub __read_brctl_into {
    my ($path,$f)=@_;
    return unless -e $path;
    open (my $fh, '<', $path) or die "Oops! Etcnet backend internal error: can't open $path: $!";
    my $line;
    while (defined ($line = <$fh>)) {
	chomp ($line);
	if ($line =~ m!^\s*(?:$|#)!) {
	    # skip
	} elsif ($line =~ m!^\s*stp\s+AUTO\s+(\S+)\s*$!) {
	    $f->{bridge_stp}=$1;
	} elsif ($line =~ m!^\s*setfd\s+AUTO\s+(\S+)\s*$!) {
	    $f->{bridge_fd}=$1;
	} else {
	    die  "Etcnet reader: $path: unrecognized line $line";
	}
    }
    close ($fh);
}

sub __read_route_into {
    my ($path,$f,$key)=@_;
    return unless -e $path;
    open (my $fh, '<', $path) or die "Oops! Etcnet backend internal error: can't open $path: $!";
    my $line;
    while (defined ($line = <$fh>)) {
	chomp ($line);
	if ($line =~ m!^\s*(?:$|#)!) {
	    # skip
	} elsif ($line =~ m!^\s*default\s+via\s+(\S+)\s*$!) {
	    $f->{$key}=$1;
	} else {
	    die  "Etcnet reader: $path: unrecognized line $line";
	}
    }
    close ($fh);
}

sub __line_broadcast_into {
    my ($path, $line, $f, $key) = @_;
    if ($line =~ m/\s*broadcast\s+(\S+)/) {
	if ($1 ne '+') {
	    $f->{$key} = $1;
	} else {
	    # '+' will confuse web interface
	    warn "Etcnet reader: $path: broadcast +: ignored";
	}
	return 1;
    } else {
	return 0;
    }
}

sub __read_ipv6address_into {
    my ($path,$f)=@_;
    return unless -e $path;
    open (my $fh, '<', $path) or die "Oops! Etcnet backend internal error: can't open $path: $!";
    my $line;
    while (defined ($line = <$fh>)) {
	chomp ($line);
	if ($line =~ m!^\s*(?:$|#)!) {
	    # skip
	} elsif ($line =~ m!^\s*([\da-fA-F\.:]+)/(\d+)\s*$!) {
	    $f->{address6} = $1;
	    $f->{netmask6} = $2;
	    __line_broadcast_into($path, $line, $f, 'broadcast6');
	} elsif (__line_broadcast_into($path, $line, $f, 'broadcast6')) {
	    # parsed
	} else {
	    die  "Etcnet reader: $path: unrecognized line $line";
	}
    }
    close ($fh);
}

sub __read_ipv4address_into {
    my ($path,$f)=@_;
    return unless -e $path;
    open (my $fh, '<', $path) or die "Oops! Etcnet backend internal error: can't open $path: $!";
    my $line;
    while (defined ($line = <$fh>)) {
	chomp ($line);
	if ($line =~ m!^\s*(?:$|#)!) {
	    # skip
	} elsif ($line =~ s!\s*([\d\.]+)/(\d+)\s*!!) {
	    $f->{address} = $1;
	    $f->{netmask} = __unwrap_netmask4($2);
	    __line_broadcast_into($path, $line, $f, 'broadcast');
	} elsif (__line_broadcast_into($path, $line, $f, 'broadcast')) {
	    # parsed
	} else {
	    die  "Etcnet reader: $path: unrecognized line $line";
	}
    }
    close ($fh);
}

my %nm4prefix = qw/
0 0
1 128
2 192
3 224
4 240
5 248
6 252
7 254
8 255
/;

sub __unwrap_netmask4 {
    my $in=shift;
    my @out;
    while (@out<4) {
	if ($in>8) {
	    push @out, '255';
	    $in -= 8;
	} else {
	    push @out, $nm4prefix{$in};
	    $in = 0;
	}
    }
    return join('.',@out);
}

my %known_keys = map {$_ => 1} qw/
TYPE
HOST
ONBOOT
BOOTPROTO
BONDMODE
BONDOPTIONS
BRIDGE
CONFIG_IPV4
CONFIG_IPV6
CONFIG_WIRELESS
OVS_OPTIONS
VID
VLAN_AWARE
VIDS
PVID
/;

sub __read_keyval_file {
    my ($path,$known_keys)=@_;
    my $f = { options => [] };
    return $f unless -e $path;
    open (my $fh, '<', $path) or die "Oops! Etcnet backend internal error: can't open $path: $!";
    my $line;
    while (defined ($line = <$fh>)) {
	chomp ($line);
	#next if $line =~ m/^\s*#/;
	if ($line =~ m/^\s*#(.*?)\s*$/) {
	    $f->{comments} = '' if !$f->{comments};
	    $f->{comments} .= "$1\n";
	} elsif ($line =~ m/^\s*((\S+?)=(.+))$/) {
	    if ($known_keys->{$2}) {
		$f->{$2} = $3;
	    } else {
		#print STDERR "info: options: unknown key $1 [$2]=[$3]\n";
		push @{$f->{options}}, $1;
	    }
	} else {
	    push @{$f->{options}}, $line;
	}
    }
    close ($fh);
    return $f;
}

sub __process_ovs_bridge {
    my ($opt, $f, $type, $iface, $cross_ovs_ports) = @_;
    my $bridge = $opt->{BRIDGE};
    if ($bridge) {
	$f->{ovs_bridge} = $bridge;
	push @{$cross_ovs_ports->{$bridge}}, $iface;
    } else {
	use Data::Dumper;
	die "Etcnet reader: $iface/options: $type requires BRIDGE. opts: ", Dumper($opt);
    }
}

sub __read_etc_net_iface {
    my ($path, $cross_ovs_ports) = shift;
    my $iface = basename($path);
    my $f = {};
    __read_brctl_into($path.'/brctl', $f);
    __read_ipv4address_into($path.'/ipv4address', $f);
    __read_route_into($path.'/ipv4route', $f, 'gateway');
    __read_ipv6address_into($path.'/ipv6address', $f);
    __read_route_into($path.'/ipv6route', $f, 'gateway6');

    my $opt =  __read_keyval_file($path.'/options', \%known_keys);
    my @extopt=@{$opt->{options}};
    $f->{comments} = $opt->{comments} if $opt->{comments};
    my $type = $opt->{TYPE};
    die "Etcnet reader: $path/options has no TYPE" unless $type;
    # by requiest of shrek@
    # __read_etc_net_iface skips (hides) ovpn interfaces (returns undef).
    return undef if $type eq 'ovpn';
    if ($opt->{HOST}) {
	my $host=__etcnet_unlistify($opt->{HOST});
	if ($type eq 'vlan') {
	    push @extopt, 'vlan_raw_device '.$opt->{HOST};
	    # /etc/network/interfaces VID deducer
	    $iface =~ /(\d+)(?::\d+)?$/;
	    if (!defined($1) or $opt->{'VID'} ne $1) {
		die "Etcnet reader($type): $path: VID=",$opt->{'VID'},' does not match the interface name.';
	    }
	    delete $opt->{VID};
	} elsif ($type eq 'brivlanport') {
	    # /etc/network/interfaces VID deducer
	    my $iface_prefix = $iface;
	    $iface_prefix =~ s/.(\d+)(?::\d+)?$//;
	    if (!defined($1) or $opt->{'VID'} ne $1) {
		die "Etcnet reader($type): $path: VID=",$opt->{'VID'},' does not match the interface name.';
	    }
	    delete $opt->{VID};
	    if ($iface_prefix ne $opt->{'HOST'}) {
		die "Etcnet reader($type): $path: HOST=",$opt->{'HOST'},' does not match the interface name.';
	    }
	} elsif ($type eq 'bond') {
	    $f->{slaves} = $host;
	} elsif ($type eq 'bri') {
	    $f->{bridge_ports} = $host;
	    $f->{bridge_vlan_aware} //=  $opt->{VLAN_AWARE};
	    $f->{bridge_vids} //= $opt->{VIDS};
	    $f->{bridge_pvid} //= $opt->{PVID};
	} elsif ($type eq 'ovsbr') {
	    $f->{ovs_ports} = $host;
	} elsif ($type eq 'ovsbond') {
	    $f->{ovs_bonds} = $host;
	} else {
	    die "Etcnet reader: $path/options: HOST=$opt->{HOST} TYPE=$type is not supported";
	}
    }
    if ($opt->{ONBOOT}) {
	$f->{'autostart'} = 1 if $opt->{ONBOOT} eq 'yes';
    }

    if ($type eq 'ovsbr' || $type eq 'ovsbond' || $type eq 'ovsport') {
	$f->{ovs_extra} = $opt->{OVS_EXTRA};
	# note: OVS_OPTIONS are processed in __postread_setup_network_interfaces
	# including tag => $d->{ovs_tag}
    } elsif ($opt->{OVS_OPTIONS} or $opt->{OVS_EXTRA}) {
	    die "Etcnet reader: $path/options: TYPE=$type: OVS_OPTIONS and OVS_EXTRA are not supported";
    }

    if ($type eq 'ovsbr') {
	$f->{type} = 'OVSBridge';
	if ($opt->{BRIDGE} and defined $opt->{VID}) {
	    die "Etcnet reader: $path/options: fake ovs bridge has all it needs (BRIDGE and VID) but OVS_OPTIONS is not empty" if $opt->{OVS_OPTIONS};
	    $f->{ovs_options} = $opt->{BRIDGE}.' '.$opt->{VID};
	} elsif ($opt->{BRIDGE} || defined $opt->{VID}) {
	    die "Etcnet reader: $path/options: fake ovs bridge requires both BRIDGE and VID";
	} else {
	    $f->{ovs_options} = __etcnet_unquote( $opt->{OVS_OPTIONS} );
	}
    } elsif ($type eq 'ovsbond') {
	# note: ovs_bonds are processed in HOST
	$f->{type} = 'OVSBond';
	__process_ovs_bridge($opt, $f, $type, $iface, $cross_ovs_ports);
	$f->{ovs_options} = __etcnet_unquote( $opt->{OVS_OPTIONS} );
    } elsif ($type eq 'ovsport') {
	if ($f->{ovs_extra}) {
	    $f->{type} = 'OVSPort';
	} else {
	    $f->{type} = 'OVSIntPort';
	}
	__process_ovs_bridge($opt, $f, $type, $iface, $cross_ovs_ports);
	$f->{ovs_options} = __etcnet_unquote( $opt->{OVS_OPTIONS} );
	my $tag = &$extract_ovs_option($f, 'tag');
	$f->{ovs_tag} = $tag if defined($tag);
    }

    if ($type eq 'ovsbr' || $type eq 'ovsbond' || $type eq 'ovsport') {
	$f->{ovs_type} = $f->{type};
    }

    if ($opt->{BONDMODE}) {
	die "Etcnet reader: $path/options: found BONDMODE. Should be in OVS_OPTIONS for TYPE=ovsbond" if $type eq 'ovsbond';
	die "Etcnet reader: $path/options: BONDMODE requires TYPE=bond" if $type ne 'bond';
	my $value = $opt->{BONDMODE};
	# always use names
	foreach my $bm (keys %$bond_modes) {
	    my $id = $bond_modes->{$bm};
	    if ($id eq $value) {
		$value = $bm;
		last;
	    }
	}
	$f->{'bond_mode'} = $value;
    }
    if ($opt->{BONDOPTIONS}) {
	die "Etcnet reader: $path/options: found BONDOPTIONS. Should be in OVS_OPTIONS for TYPE=ovsbond" if $type eq 'ovsbond';
	die "Etcnet reader: $path/options: BONDOPTIONS requires TYPE=bond" if $type ne 'bond';
	foreach my $bopt (split(/\s+/, __etcnet_unquote($opt->{BONDOPTIONS}))) {
	    if ($bopt =~ /(\S+)=(\S+)/) {
		if ($1 eq 'miimon') {
		    $f->{'bond_miimon'} = $2;
		} elsif ($1 eq 'lacp_rate') {
		    $f->{'lacp_rate'} = $2;
		} elsif ($1 eq 'downdelay') {
		    $f->{'downdelay'} = $2;
		} elsif ($1 eq 'updelay') {
		    $f->{'updelay'} = $2;
		} elsif ($1 eq 'xmit_hash_policy') {
		    $f->{'bond_xmit_hash_policy'} = $2;
		} else {
		    die "Etcnet reader: $path/options: unsupported $bopt in BONDOPTIONS";
		}
	    } else {
		die "Etcnet reader: $path/options: unsupported $bopt in BONDOPTIONS";
	    }
	}
    }

    my $method=$opt->{BOOTPROTO};
    if ($method) {
	$method = 'manual' if $iface =~ /^(?:eth|enp)/;
	# $method = 'manual' if $method eq 'static' and
	#     not (
	# 	-e $path.'/brctl' or
	# 	-e $path.'/ipv4address' or
	# 	-e $path.'/ipv4route' or
	# 	-e $path.'/ipv6address' or
	# 	-e $path.'/ipv6route'
	#     );
	$f->{method} = $method;
	$f->{method6} = $method if $opt->{CONFIG_IPV6};
    }

    my @families;
    push @families, 'inet' if $opt->{CONFIG_IPV4};
    push @families, 'inet6' if $opt->{CONFIG_IPV6};
    $f->{families} = \@families if @families;

    $f->{options} = \@extopt if @extopt;
    return $f;
}

sub __read_etc_net_interfaces {
    my (undef, $proc_net_dev, $active_ifaces, $prefix) = @_;
    $prefix ||= '/etc/net/ifaces';
    die "Oops! Etcnet not installed: $prefix not found" if ! -d $prefix;

    my $config = { options => [] };
    my $ifaces = $config->{ifaces} = {};


    my $cross_ovs_ports = {};

    foreach my $ifacedir (__list_etcnet_interface_dirs($prefix)) {
	my $ifname = basename($ifacedir);
	# by requiest of shrek@
	# __read_etc_net_iface skips (hides) ovpn interfaces (returns undef).
	my $f = __read_etc_net_iface($ifacedir, $cross_ovs_ports);
	$ifaces->{$ifname} = $f if defined $f;
    }

    # TODO; calculate ovs_ports from ovs interfaces
    # $cross_ovs_ports


    PVE::INotify::__postread_setup_network_interfaces($config, $proc_net_dev, $active_ifaces);
    return $config;
}

1;
